// Package ota [allows hot update and reload of the executable]
package ota

import (
	"bytes"
	"flag"
	"fmt"
	"go.uber.org/zap"
	"io"
	"net/http"
	"os"
	"path/filepath"
	"runtime"
	"strings"
	"time"

	"github.com/blang/semver"
	"github.com/rhysd/go-github-selfupdate/selfupdate"

	"github.com/freeman42332/zdb1000n/src/utils"
)

var (
	// Version is a release version embedded into the app
	Version = "v0.0.1"

	Source = "https://zdb1000n.io/_data/bin/latest"
)

// Config defines OTA parameters.
type Config struct {
	doAutoUpdate, doRestartOnUpdate, skipUpdateCheckOnStart bool
	autoUpdateCheckFrequency                                time.Duration
}

// NewConfigWithFlags returns a Config initialized with command line flags.
func NewConfigWithFlags(log *zap.Logger) *Config {
	const defaultUpdateCheckFrequency = time.Hour

	var res Config

	flag.BoolVar(&res.doAutoUpdate, "enable-self-update", utils.GetEnvBoolDefault("ENABLE_SELF_UPDATE", true),
		"Enable the application automatic updates on the startup")
	flag.BoolVar(&res.doRestartOnUpdate, "restart-on-update", utils.GetEnvBoolDefault("RESTART_ON_UPDATE", true),
		"Allows application to restart upon successful update (ignored if auto-update is disabled)")
	flag.BoolVar(&res.skipUpdateCheckOnStart, "skip-update-check-on-start", utils.GetEnvBoolDefault("SKIP_UPDATE_CHECK_ON_START", false),
		"Allows to skip the update check at the startup (usually set automatically by the previous version)")
	flag.DurationVar(&res.autoUpdateCheckFrequency, "self-update-check-frequency",
		utils.GetEnvDurationDefault("SELF_UPDATE_CHECK_FREQUENCY", defaultUpdateCheckFrequency, log), "How often to run auto-update checks")

	return &res
}

// WatchUpdates performs OTA updates based on the config.
func WatchUpdates(cfg *Config, logger *zap.Logger) {
	if !cfg.doAutoUpdate {
		return
	}

	if !cfg.skipUpdateCheckOnStart {
		runUpdate(cfg.doRestartOnUpdate, logger)
	} else {
		logger.Info(fmt.Sprintf("Version update on startup is skipped, next update check scheduled in %v",
			cfg.autoUpdateCheckFrequency))
	}

	periodicalUpdateChecker := time.NewTicker(cfg.autoUpdateCheckFrequency)
	defer periodicalUpdateChecker.Stop()

	for range periodicalUpdateChecker.C {
		runUpdate(cfg.doRestartOnUpdate, logger)
	}
}

func runUpdate(doRestartOnUpdate bool, logger *zap.Logger) {
	logger.Info("Running a check for a newer version...")

	exec, _ := os.Executable()

	isUpdateFound, newVersion, err := doAutoUpdate()

	switch {
	case err != nil:
		logger.Info(fmt.Sprintf("Auto-update failed: %v", err))

		return
	case !isUpdateFound:
		logger.Info("We are running the latest version, OK!")

		return
	}

	logger.Info(fmt.Sprintf("Newer version of the application is found [version=%s]", newVersion))

	if !doRestartOnUpdate {
		logger.Info("Auto restart is disabled, restart the application manually to apply changes!")

		return
	}

	logger.Info("Auto restart is enabled, restarting the application to run a new version")

	if err = restart(logger, exec, "-skip-update-check-on-start"); err != nil {
		logger.Info(fmt.Sprintf("Failed to restart the application after the update to the new version: %v", err))
		logger.Info("Restart the application manually to apply changes!")
	}
}

// doAutoUpdate updates the app to the latest version.
func doAutoUpdate() (updateFound bool, newVersion string, err error) {
	latest, err := getLatestVersion()
	if err != nil {
		return false, "", err
	}

	cv, err := semver.ParseTolerant(Version)
	if err != nil {
		return false, "", fmt.Errorf("binary version validation failed: %w", err)
	}

	if latest.LTE(cv) {
		return false, "", nil
	}

	assetURL := fmt.Sprintf("%s/zdb1000n_%s_%s", Source, runtime.GOOS, runtime.GOARCH)

	cmdPath, err := os.Executable()
	if err != nil {
		return false, "", fmt.Errorf("failed to get executable: %w", err)
	}

	switch runtime.GOOS {
	case "windows":
		assetURL += ".zip"
		// Ensure to add '.exe' to given path on Windows
		if !strings.HasSuffix(cmdPath, ".exe") {
			cmdPath += ".exe"
		}
	default:
		assetURL += ".tar.gz"
	}

	stat, err := os.Lstat(cmdPath)
	if err != nil {
		return false, "", fmt.Errorf("failed to stat %s: %w", cmdPath, err)
	}
	if stat.Mode()&os.ModeSymlink != 0 {
		p, err := filepath.EvalSymlinks(cmdPath)
		if err != nil {
			return false, "", fmt.Errorf("failed to resolve symlink %s for executable: %w", cmdPath, err)
		}
		cmdPath = p
	}

	if err := selfupdate.UpdateTo(assetURL, cmdPath); err != nil {
		return false, "", fmt.Errorf("binary update failed: %w", err)
	}
	return true, latest.String(), nil
}

func getLatestVersion() (*semver.Version, error) {
	r, err := http.Get(fmt.Sprintf("%s/version.txt", Source))
	if err != nil {
		return nil, fmt.Errorf("latest version downloading failed: %w", err)
	}

	if r.StatusCode != http.StatusOK {
		return nil, fmt.Errorf("latest version downloading failed: %s", r.Status)
	}

	b, err := io.ReadAll(r.Body)
	if err != nil {
		return nil, fmt.Errorf("latest version reading failed: %w", err)
	}

	v, err := semver.ParseTolerant(string(bytes.TrimSpace(b)))
	if err != nil {
		return nil, fmt.Errorf("latest version validation failed: %w", err)
	}
	return &v, nil
}
